@@ -12,12 +12,9 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-from __future__ import absolute_import
+__version__ = "2.0.0"
 
-
-__version__ = '0.5.6'
-
-DEFAULT_CHARSET = 'UTF-8'
+DEFAULT_CHARSET = "UTF-8"
 
 from .request import *  # noqa
 from .requestcontext import context  # noqa
@@ -25,5 +22,6 @@ from .response import *  # noqa
 from .core import *  # noqa
 from .routing import *  # noqa
 from .routeargs import *  # noqa
+from .subrequests import *  # noqa
 
 from .util.common import *  # noqa
