@@ -12,29 +12,38 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-from __future__ import absolute_import
-
-import sys
-
-from collections import defaultdict
-from contextlib import contextmanager
 from functools import partial
+from typing import Callable
+from typing import Dict
+from typing import List
+from typing import Tuple
+from typing import Type
+from typing import Set
+from typing import Union
+import contextlib
 import logging
+import sys
+import types
 
-from fresco.compat import reraise
 from fresco.request import Request
 from fresco.response import Response
 from fresco.util.http import encode_multipart
 from fresco.util.urls import normpath, make_query
 from fresco.util.common import fq_path
+from fresco.util.wsgi import make_environ
 
 from fresco.exceptions import ResponseException
 from fresco.requestcontext import context
-from fresco.routing import (GET, ExtensiblePattern, RouteCollection,
-                            RouteNotFound, TraversedCollection)
+from fresco.routing import (
+    GET,
+    ExtensiblePattern,
+    RouteCollection,
+    RouteNotFound,
+    TraversedCollection,
+)
 from fresco.options import Options
 
-__all__ = ('FrescoApp', 'urlfor')
+__all__ = ("FrescoApp", "urlfor")
 
 logger = logging.getLogger(__name__)
 
@@ -55,23 +64,17 @@ class FrescoApp(RouteCollection):
 
     def __init__(self, *args, **kwargs):
 
-        from fresco.middleware import FrescoCoreMiddleware
-
-        views = kwargs.pop('views', None)
-        path = kwargs.pop('path', None)
+        views = kwargs.pop("views", None)
+        path = kwargs.pop("path", None)
         super(FrescoApp, self).__init__(*args, **kwargs)
 
         if views is not None:
             if path is None:
-                path = '/'
+                path = "/"
             self.include(path, views)
 
         #: A list of (middleware, args, kwargs) tuples
-        self._middleware = []
-
-        #: Middleware layers applied after any middleware added through
-        #: :meth:`add_middleware`.
-        self.core_middleware = [partial(FrescoCoreMiddleware, frescoapp=self)]
+        self._middleware: List[Tuple[Callable, Tuple, Dict]] = []
 
         #: The WSGI application. This is generated when the first request is
         #: made.
@@ -85,7 +88,7 @@ class FrescoApp(RouteCollection):
         #: Each function will be passed the request object.
         #: If a function returns a value (other than ``None``)
         #: this will be returned and the normal routing system bypassed.
-        self.process_request_handlers = []
+        self.process_request_handlers: List[Callable] = []
 
         #: Functions to be called after routing, but before any view is invoked
         #: Each function will be passed
@@ -95,14 +98,14 @@ class FrescoApp(RouteCollection):
         #: instead of calling the scheduled view.
         #: If a function returns any other (non-None) value this will be used
         #: to replace the scheduled view function.
-        self.process_view_handlers = []
+        self.process_view_handlers: List[Callable] = []
 
         #: Functions to be called after a response object has been generated
         #: Each function will be passed ``request, response``.
         #: If a function returns a value (other than ``None``),
         #: this value will be
         #: returned as the response instead of calling the scheduled view.
-        self.process_response_handers = []
+        self.process_response_handlers: List[Callable] = []
 
         #: Functions to be called if the response has an HTTP error status code
         #: (400 <= x <= 599)
@@ -110,7 +113,9 @@ class FrescoApp(RouteCollection):
         #: If a function returns a value (other than ``None``),
         #: this value will be
         #: returned as the response instead of calling the scheduled view.
-        self.process_http_error_response_handlers = []
+        self.process_http_error_response_handlers: List[
+            Tuple[int, Callable]
+        ] = []
 
         #: Functions to be called if an exception is raised during a view
         #: Each function will be passed ``request, exc_info``.
@@ -118,20 +123,20 @@ class FrescoApp(RouteCollection):
         #: this value will be
         #: returned as the response and the error will not be propagated.
         #: If all exception handlers return None then the error will be raised
-        self.process_exception_handlers = []
+        self.process_exception_handlers: List[Tuple[Exception, Callable]] = []
 
         #: Functions to be called at the end of request processing,
         #: after all content has been output.
         #: Each function will be passed ``request`` and should not
         #: return any value.
-        self.process_teardown_handlers = []
+        self.process_teardown_handlers: List[Callable] = []
 
     def __call__(self, environ, start_response):
         """\
         Call the app as a WSGI application
         """
         if self._wsgi_app is None:
-            self._wsgi_app = self._make_wsgi_app()
+            self._wsgi_app = self.make_wsgi_app()
         return self._wsgi_app(environ, start_response)
 
     def __str__(self):
@@ -139,28 +144,40 @@ class FrescoApp(RouteCollection):
         String representation of the application and its configured routes
         """
         clsname = self.__class__.__name__
-        return '<%s %s>' % (
+        return "<%s %s>" % (
             clsname,
-            ('\n' + ' ' * (len(clsname) + 2)).join(str(r)
-                                                   for r in self.__routes__))
+            ("\n" + " " * (len(clsname) + 2)).join(
+                str(r) for r in self.__routes__
+            ),
+        )
 
     def get_response(
-            self, routecollection=None, request=None, path=None, method=None):
+        self,
+        routecollection=None,
+        request=None,
+        path=None,
+        method=None,
+        currentcontext=context.currentcontext,
+        normpath=normpath,
+    ):
+
+        ctx = currentcontext()
         try:
             routecollection = routecollection or self
-            request = request or context.request
+            request = request or ctx["request"]
             method = method or request.method
             path = path or request.path_info
             error_response = None
 
-            context.app = self
+            ctx["app"] = self
             environ = request.environ
-            environ['fresco.app'] = self
+            environ["fresco.app"] = self
 
             if path:
-                path = normpath(path)
+                if ".." in path or "//" in path or "/./" in path:
+                    path = normpath(path)
             else:
-                path = '/'
+                path = "/"
         except ResponseException as e:
             return e.response
 
@@ -171,61 +188,87 @@ class FrescoApp(RouteCollection):
                 if r is not None:
                     response = r
             except Exception:
-                return self.handle_exception(request,
-                                             allow_reraise=False)
+                return self.handle_exception(request, allow_reraise=False)
         if response:
             return response
 
-        for traversal in routecollection.get_routes(path, method, request):
-            try:
-                route = traversal.route
-                environ['wsgiorg.routing_args'] = (traversal.args,
-                                                   traversal.kwargs)
-                view = route.getview(method)
-                context.view_self = getattr(view, '__self__', None)
+        try:
+            for traversal in routecollection.get_routes(path, method, request):
+                try:
+                    route = traversal.route
+                    environ["wsgiorg.routing_args"] = (
+                        traversal.args,
+                        traversal.kwargs,
+                    )
+                    view = route.getview(method)
+                    ctx["view_self"] = getattr(view, "__self__", None)
+                    ctx["route_traversal"] = traversal
+                    if self.logger:
+                        self.logger.info(
+                            "matched route: %s %r => %r",
+                            request.method,
+                            path,
+                            fq_path(view),
+                        )
 
-                if self.logger:
-                    self.logger.info("matched route: %s %r => %r",
-                                     request.method, path, fq_path(view))
+                    response = None
+                    for f in self.process_view_handlers:
+                        try:
+                            r = f(
+                                request, view, traversal.args, traversal.kwargs
+                            )
+                            if r is not None:
+                                response = r
+                        except Exception:
+                            return self.handle_exception(
+                                request, allow_reraise=False
+                            )
+                    if response is not None:
+                        if isinstance(response, Response):
+                            return response
+                        else:
+                            view = response
 
-                context.route_traversal = traversal
+                    view = route.getdecoratedview(view)
+                    response = view(*traversal.args, **traversal.kwargs)
+                    if (
+                        route.fallthrough_statuses
+                        and response.status_code in route.fallthrough_statuses
+                    ):
+                        continue
 
-                response = None
-                for f in self.process_view_handlers:
-                    try:
-                        r = f(request, view, traversal.args, traversal.kwargs)
-                        if r is not None:
-                            response = r
-                    except Exception:
-                        return self.handle_exception(request,
-                                                     allow_reraise=False)
-                if response is not None:
-                    if isinstance(response, Response):
-                        return response
-                    else:
-                        view = response
+                except ResponseException as e:
+                    if e.is_final:
+                        return e.response
+                    error_response = error_response or e.response
+                    if (
+                        route.fallthrough_statuses
+                        and error_response.status_code
+                        in route.fallthrough_statuses
+                    ):
+                        continue
 
-                view = route.getdecoratedview(view)
-                response = view(*traversal.args, **traversal.kwargs)
+                except Exception:
+                    return self.handle_exception(request)
+                else:
+                    return response
 
-            except ResponseException as e:
-                if e.is_final:
-                    return e.response
-                error_response = error_response or e.response
+        except ResponseException as e:
+            if e.is_final:
+                return e.response
+            error_response = error_response or e.response
 
-            except Exception:
-                return self.handle_exception(request)
-            else:
-                return response
+        except Exception:
+            return self.handle_exception(request)
 
         # A route was matched, but an error was returned
         if error_response:
             return error_response
 
         # Is this a head request?
-        if method == 'HEAD':
+        if method == "HEAD":
             response = self.get_response(routecollection, request, path, GET)
-            if '200' <= response.status < '300':
+            if "200" <= response.status < "300":
                 return response.replace(content=[], content_length=0)
             return response
 
@@ -235,9 +278,9 @@ class FrescoApp(RouteCollection):
             return Response.method_not_allowed(methods)
 
         # Is the URL just missing a trailing '/'?
-        if path[-1] != '/':
-            for _ in self.get_methods(routecollection, request, path + '/'):
-                return Response.redirect_permanent(path + '/')
+        if path[-1] != "/":
+            for _ in self.get_methods(routecollection, request, path + "/"):
+                return Response.unrestricted_redirect_permanent(path + "/")
 
         return Response.not_found()
 
@@ -245,7 +288,7 @@ class FrescoApp(RouteCollection):
         request = request or context.request
         response = self.get_response(request=request)
 
-        for f in self.process_response_handers:
+        for f in self.process_response_handlers:
             try:
                 r = f(request, response)
                 if r is not None:
@@ -253,8 +296,9 @@ class FrescoApp(RouteCollection):
             except Exception:
                 self.log_exception(request)
 
-        if '400' <= response.status <= '599':
-            response = self.handle_http_error_response(request, response)
+        if "400" <= response.status <= "599":
+            if self.process_http_error_response_handlers:
+                response = self.handle_http_error_response(request, response)
 
         return response
 
@@ -278,7 +322,7 @@ class FrescoApp(RouteCollection):
         """\
         Return the HTTP methods valid in routes to the given path
         """
-        methods = set()
+        methods: Set[str] = set()
         for traversal in routecollection.get_routes(path, None):
             route = traversal.route
             if route.predicate and not route.predicate(request):
@@ -290,41 +334,53 @@ class FrescoApp(RouteCollection):
         exc_info = exc_info or sys.exc_info()
         (self.logger or logger).error(
             "Exception in {0} {1}".format(request.method, request.url),
-            exc_info=exc_info)
+            exc_info=exc_info,
+        )
 
-    def handle_exception(self, request, allow_reraise=True):
+    def handle_exception(
+        self, request, allow_reraise=True
+    ) -> Union[
+        Response,
+        Tuple[Type[BaseException], BaseException, types.TracebackType],
+    ]:
 
         exc_info = sys.exc_info()
+        if exc_info[0] is None:
+            raise AssertionError(
+                "handle_exception called " "when no exception is being handled"
+            )
 
         have_error_handlers = self.process_exception_handlers or any(
             st in (None, 500)
-            for st, fn in self.process_http_error_response_handlers)
+            for st, fn in self.process_http_error_response_handlers
+        )
 
         # Backwards compatibility: if no exception or 500 http error
         # handlers have been installed we default to the old behavior
         # of raising the exception and letting the upstream
         # server handle it
         if allow_reraise and not have_error_handlers:
-            reraise(exc_info)
+            raise exc_info[1].with_traceback(exc_info[2])  # type: ignore
+        response: Union[
+            Response,
+            Tuple[Type[BaseException], BaseException, types.TracebackType],
+        ] = Response.internal_server_error()
 
-        try:
+        if not self.process_exception_handlers:
             self.log_exception(request, exc_info)
-            response = None
-            for exc_type, func in self.process_exception_handlers:
-                try:
-                    if not issubclass(exc_info[0], exc_type):
-                        continue
-                    r = func(request, exc_info)
-                    if r is not None:
-                        response = r
-                except Exception:
-                    self.log_exception(request)
-            if response is not None:
-                return response
-            else:
-                return Response.internal_server_error()
-        finally:
-            del exc_info
+        for exc_type, func in self.process_exception_handlers:
+            try:
+                if not issubclass(exc_info[0], exc_type):  # type: ignore
+                    continue
+                r = func(request, exc_info)
+                if r is not None:
+                    response = r
+                    break
+            except Exception:
+                self.log_exception(request)
+        if isinstance(response, tuple) and len(response) == 3:
+            raise exc_info[1].with_traceback(exc_info[2])  # type: ignore
+        return response
 
     def add_middleware(self, middleware, *args, **kwargs):
         """\
@@ -334,7 +390,7 @@ class FrescoApp(RouteCollection):
         middleware added will occupy the innermost layer and be called last in
         each request cycle.
         """
-        self._wsgi_app = None
+        self.reset_wsgi_app()
         self._middleware.append((middleware, args, kwargs))
 
     def remove_middleware(self, middleware):
@@ -345,9 +401,10 @@ class FrescoApp(RouteCollection):
                            that was previouslly passed to
                            :meth:`add_middleware`.
         """
-        self._wsgi_app = None
-        self._middleware = [m for m in self._middleware
-                            if m[0] is not middleware]
+        self.reset_wsgi_app()
+        self._middleware = [
+            m for m in self._middleware if m[0] is not middleware
+        ]
 
     def insert_middleware(self, position, middleware, *args, **kwargs):
         """
@@ -363,21 +420,89 @@ class FrescoApp(RouteCollection):
         :param kwargs: extra keyword args to be passed to the middleware for
                        initialization
         """
-        self._wsgi_app = None
+        self.reset_wsgi_app()
         self._middleware.insert(position, (middleware, args, kwargs))
 
-    def _raw_wsgi_app(self, environ, start_response):
-        request = self.request_class(environ)
-        response = self.view(request)
-        return response(environ, start_response)
+    def make_wsgi_app(self, wsgi_app=None, use_middleware=True):
+        """
+        Return a WSGI (PEP-3333) compliant application that drives this
+        FrescoApp object.
 
-    def _make_wsgi_app(self, app=None):
-        app = app or self._raw_wsgi_app
-        for middleware, args, kwargs in self._middleware:
-            app = middleware(app, *args, **kwargs)
-        for middleware in self.core_middleware:
-            app = middleware(app)
-        return app
+        :param wsgi_app: if given, will be called in place of
+                    :meth:~`fresco.core.FrescoApp.view`(request)
+                    with :meth:~`fresco.core.FrescoApp.requestcontext`)
+        :param use_middleware: if True, the app's middleware stack will be applied
+                            to the resulting WSGI app.
+        """
+        if wsgi_app is None:
+
+            def wsgi_app(
+                environ,
+                start_response,
+                view=self.view,
+                request_class=self.request_class,
+            ):
+                request = request_class(environ)
+                return view(request)(environ, start_response)
+
+        if use_middleware:
+            for m, m_args, m_kwargs in self._middleware:
+                wsgi_app = m(wsgi_app, *m_args, **m_kwargs)
+
+        def fresco_wsgi_app(
+            environ,
+            start_response,
+            frescoapp=self,
+            wsgi_app=wsgi_app,
+            request_class=self.request_class,
+            process_teardown_handlers=self.process_teardown_handlers,
+            call_process_teardown_handlers=self.call_process_teardown_handlers,
+            context_push=context.push,
+            context_pop=context.pop,
+        ):
+            request = request_class(environ)
+            context_push(request=request, app=frescoapp)
+            iterator = None
+            try:
+                iterator = wsgi_app(environ, start_response)
+                for item in iterator:
+                    yield item
+            except Exception:
+                exc_info = sys.exc_info()
+                try:
+                    response = frescoapp.handle_exception(request)
+                    if "400" <= response.status <= "599":
+                        response = frescoapp.handle_http_error_response(
+                            request, response
+                        )
+
+                    def exc_start_response(s, h, exc_info=exc_info):
+                        return start_response(s, h, exc_info)
+
+                    for item in response(environ, exc_start_response):
+                        yield item
+                finally:
+                    del exc_info
+            finally:
+                try:
+                    if process_teardown_handlers:
+                        call_process_teardown_handlers(request)
+                finally:
+                    try:
+                        close = getattr(iterator, "close", None)
+                        if close is not None:
+                            close()
+                    finally:
+                        context_pop()
+
+        return fresco_wsgi_app
+
+    def reset_wsgi_app(self):
+        try:
+            del self.__call__
+        except AttributeError:
+            pass
+        self._wsgi_app = None
 
     def urlfor(self, viewspec, *args, **kwargs):
         """\
@@ -396,34 +521,40 @@ class FrescoApp(RouteCollection):
         All other arguments or keyword args are fed to the ``pathfor`` method
         of the pattern.
         """
-        scheme = kwargs.pop('_scheme', None)
-        netloc = kwargs.pop('_netloc', None)
-        query = kwargs.pop('_query', {})
-        script_name = kwargs.pop('_script_name', None)
-        fragment = kwargs.pop('_fragment', None)
+        scheme = kwargs.pop("_scheme", None)
+        netloc = kwargs.pop("_netloc", None)
+        query = kwargs.pop("_query", {})
+        script_name = kwargs.pop("_script_name", None)
+        fragment = kwargs.pop("_fragment", None)
 
-        request = context.request
-        traversal = context.get('route_traversal')
+        ctx = context.currentcontext()
+        request = ctx["request"]
+        traversal = ctx.get("route_traversal")
         if traversal and traversal.collections_traversed:
             collections_traversed = traversal.collections_traversed
         else:
             collections_traversed = [
-                TraversedCollection(self, '', None, (), {}, (), {})]
+                TraversedCollection(self, "", None, (), {}, (), {})
+            ]
 
         exc = None
         for ct in collections_traversed[::-1]:
             try:
-                path = ct.path + \
-                        ct.collection.pathfor(viewspec, request=request,
-                                           *args, **kwargs)
+                path = ct.path + ct.collection.pathfor(
+                    viewspec, request=request, *args, **kwargs
+                )
             except RouteNotFound as e:
                 exc = e
                 continue
-            return request.make_url(scheme=scheme, netloc=netloc,
-                                    SCRIPT_NAME=script_name,
-                                    PATH_INFO=path,
-                                    parameters='',
-                                    query=query, fragment=fragment)
+            return request.make_url(
+                scheme=scheme,
+                netloc=netloc,
+                SCRIPT_NAME=script_name,
+                PATH_INFO=path,
+                parameters="",
+                query=query,
+                fragment=fragment,
+            )
         raise exc or RouteNotFound(viewspec)
 
     def view_method_not_found(self, valid_methods):
@@ -463,7 +594,7 @@ class FrescoApp(RouteCollection):
         """
         Register a ``process_response`` hook function
         """
-        self.process_response_handers.append(func)
+        self.process_response_handlers.append(func)
         return func
 
     def process_exception(self, func, exc_type=Exception):
@@ -491,13 +622,15 @@ class FrescoApp(RouteCollection):
         Register a ``process_teardown`` hook function
         """
         if self._wsgi_app:
-            raise AssertionError("Cannot add hook: "
-                                 "application is now receiving requests")
+            raise AssertionError(
+                "Cannot add hook: application is now receiving requests"
+            )
         self.process_teardown_handlers.append(func)
 
-    @contextmanager
-    def requestcontext(self, url='/', environ=None,
-                       wsgi_input=b'', middleware=True, **kwargs):
+    @contextlib.contextmanager
+    def requestcontext(
+        self, url="/", environ=None, wsgi_input=b"", middleware=True, **kwargs
+    ):
         """
         Return the global :class:`fresco.requestcontext.RequestContext`
         instance, populated with a new request object modelling default
@@ -526,26 +659,30 @@ class FrescoApp(RouteCollection):
                        WSGI request environment
         """
 
-        from fresco.util.wsgi import WSGIStack
-        from fresco.util.wsgi import make_environ
+        def fake_app(environ, start_response):
+            start_response("200 OK", [])
+            yield b""
 
-        if middleware:
-            middleware = [
-                lambda app, m=m, a=mw_args, k=mw_kwargs: m(app, *a, **k)
-                for m, mw_args, mw_kwargs in self._middleware]
+        def fake_start_response(status, headers, exc_info=None):
+            return lambda s: None
 
-        else:
-            middleware = []
-
-        middleware.extend(self.core_middleware)
-        kwargs.update(middleware=middleware)
-        with WSGIStack(make_environ(url, environ, wsgi_input, **kwargs),
-                       middleware):
-            yield context
+        environ = make_environ(url, environ, wsgi_input, **kwargs)
+        app = self.make_wsgi_app(wsgi_app=fake_app, use_middleware=middleware)
+        content_iterator = app(environ, fake_start_response)
+        next(content_iterator, None)
+        yield context
+        list(content_iterator)
+        getattr(content_iterator, "close", lambda: None)()
 
     def requestcontext_with_payload(
-            self, url='/', data=None, environ=None, files=None,
-            multipart=False, **kwargs):
+        self,
+        url="/",
+        data=None,
+        environ=None,
+        files=None,
+        multipart=False,
+        **kwargs
+    ):
 
         if files:
             multipart = True
@@ -553,48 +690,40 @@ class FrescoApp(RouteCollection):
         if multipart:
             wsgi_input, headers = encode_multipart(data, files)
             kwargs.update(headers)
-        elif hasattr(data, 'read'):
+        elif hasattr(data, "read"):
             wsgi_input = data.read()
         elif isinstance(data, bytes):
             wsgi_input = data
         else:
-            wsgi_input = make_query(data).encode('ascii')
+            wsgi_input = make_query(data).encode("ascii")
 
-        if 'CONTENT_LENGTH' not in kwargs:
-            kwargs['CONTENT_LENGTH'] = str(len(wsgi_input))
+        if "CONTENT_LENGTH" not in kwargs:
+            kwargs["CONTENT_LENGTH"] = str(len(wsgi_input))
 
         return self.requestcontext(
-            url, environ, wsgi_input=wsgi_input, **kwargs)
+            url, environ, wsgi_input=wsgi_input, **kwargs
+        )
 
     def requestcontext_post(self, *args, **kwargs):
         return self.requestcontext_with_payload(
-            REQUEST_METHOD='POST', *args, **kwargs)
+            REQUEST_METHOD="POST", *args, **kwargs
+        )
 
     def requestcontext_put(self, *args, **kwargs):
-        kwargs['REQUEST_METHOD'] = 'PUT'
+        kwargs["REQUEST_METHOD"] = "PUT"
         return self.requestcontext_with_payload(*args, **kwargs)
 
     def requestcontext_patch(self, *args, **kwargs):
-        kwargs['REQUEST_METHOD'] = 'PATCH'
+        kwargs["REQUEST_METHOD"] = "PATCH"
         return self.requestcontext_with_payload(*args, **kwargs)
 
     def requestcontext_delete(self, *args, **kwargs):
-        kwargs['REQUEST_METHOD'] = 'DELETE'
+        kwargs["REQUEST_METHOD"] = "DELETE"
         return self.requestcontext(*args, **kwargs)
 
 
-def collect_keys(items):
-    """
-    [(k1, v1), (k1, v2), (k2, v3)] -> [(k1, [v1, v2]), (k2, [v3])]
-    """
-    d = defaultdict(list)
-    for key, value in items:
-        d[key].append(value)
-    return d.items()
-
-
 def urlfor(*args, **kwargs):
-    """\
+    """
     Convenience wrapper around :meth:`~fresco.core.FrescoApp.urlfor`.
     """
     return context.app.urlfor(*args, **kwargs)
