@@ -12,12 +12,13 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-from __future__ import absolute_import
-
 from collections import defaultdict
-from fresco.compat import get_ident
+from threading import get_ident
+from typing import Any
+from typing import Dict
+from typing import List
 
-__all__ = 'context', 'RequestContext'
+__all__ = "context", "RequestContext"
 
 
 class RequestContext(object):
@@ -27,7 +28,7 @@ class RequestContext(object):
 
     Requires an ``ident_func`` that returns a hashable identifier that will
     give a different value per request. In a threaded environment this would be
-    ``thread.get_ident``. If using a different execution model, a different
+    ``threading.get_ident``. If using a different execution model, a different
     strategy would need to be found.
 
     During a request the context object will contain the following keys:
@@ -39,10 +40,13 @@ class RequestContext(object):
     - ``request``: the current :class:`~fresco.request.Request` object.
     """
 
-    def __init__(self, ident_func=get_ident):
-        object.__setattr__(self, '_contexts', defaultdict(list))
-        object.__setattr__(self, '_ident_func', ident_func)
-        self.push()
+    __slots__ = ["_contexts", "_ident_func"]
+
+    def __init__(self, ident_func=get_ident, setattr=object.__setattr__):
+        c: Dict[Any, List] = defaultdict(list)
+        c[ident_func()] = [{}]
+        setattr(self, "_contexts", c)
+        setattr(self, "_ident_func", ident_func)
 
     def push(self, **bindnames):
         self._contexts[self._ident_func()].append(bindnames)
@@ -58,16 +62,16 @@ class RequestContext(object):
 
     def __getattr__(self, item):
         try:
-            return self.currentcontext()[item]
+            return self._contexts[self._ident_func()][-1][item]
         except (KeyError, IndexError):
             raise AttributeError(item)
 
     def __setattr__(self, item, ob):
-        self.currentcontext()[item] = ob
+        self._contexts[self._ident_func()][-1][item] = ob
 
     def __delattr__(self, item):
         try:
-            del self.currentcontext()[item]
+            del self._contexts[self._ident_func()][-1][item]
         except KeyError:
             raise AttributeError(item)
 
@@ -81,10 +85,11 @@ class RequestContext(object):
 
     def __repr__(self):
         return "<%s localdepth=%d; total=%d; current=%r>" % (
-                            self.__class__.__name__,
-                            len(self._contexts[self._ident_func()]),
-                            sum(len(ctx) for ctx in self._contexts.values()),
-                            self.currentcontext())
+            self.__class__.__name__,
+            len(self._contexts[self._ident_func()]),
+            sum(len(ctx) for ctx in self._contexts.values()),
+            self.currentcontext(),
+        )
 
 
 #: The Context for the current request; allows apps to access the current
