@@ -16,15 +16,19 @@
 Utilities functions for URL building and manipulation
 """
 
-from __future__ import absolute_import
-
+from functools import lru_cache
+from typing import Iterable
+from typing import List
+from urllib.parse import quote_plus
+from urllib.parse import urlparse
+from urllib.parse import urlunparse
 import posixpath
-from fresco.compat import quote_plus, urlparse, urlunparse, ustr, string_types
+import re
 
 import fresco
 from fresco.multidict import MultiDict
 
-__all__ = 'join_path', 'url_join', 'strip_trailing_slashes', 'normpath'
+__all__ = "join_path", "url_join", "strip_trailing_slashes", "normpath"
 
 
 def join_path(a, b):
@@ -54,13 +58,13 @@ def join_path(a, b):
     if not b:
         return a
 
-    while a and a[-1] == '/':
+    while a and a[-1] == "/":
         a = a[:-1]
 
-    while b and b[0] == '/':
+    while b and b[0] == "/":
         b = b[1:]
 
-    return a + '/' + b
+    return a + "/" + b
 
 
 def url_join(base, rel):
@@ -88,17 +92,20 @@ def url_join(base, rel):
     pbase = urlparse(base)
     path = pbase.path
     if prel.path:
-        path = normpath(posixpath.join(posixpath.dirname(pbase.path),
-                                       prel.path))
+        path = normpath(
+            posixpath.join(posixpath.dirname(pbase.path), prel.path)
+        )
 
-    return urlunparse((
-        pbase.scheme,
-        pbase.netloc,
-        path,
-        prel.params,
-        prel.query,
-        prel.fragment
-    ))
+    return urlunparse(
+        (
+            pbase.scheme,
+            pbase.netloc,
+            path,
+            prel.params,
+            prel.query,
+            prel.fragment,
+        )
+    )
 
 
 def strip_trailing_slashes(path):
@@ -108,11 +115,12 @@ def strip_trailing_slashes(path):
     :param path: a path, eg ``'/foo/bar/'``
     :return: The path with trailing slashes removed, eg ``'/foo/bar'``
     """
-    while path and path[-1] == '/':
+    while path and path[-1] == "/":
         path = path[:-1]
     return path
 
 
+@lru_cache(50)
 def normpath(path):
     """
     Return ``path`` normalized to remove '../' etc.
@@ -133,33 +141,35 @@ def normpath(path):
         >>> normpath('/etc/passwd/')
         '/etc/passwd/'
     """
-    segments = path.split('/')
-    newpath = []
+    segments = path.split("/")
+    newpath: List[str] = []
+    newpath_append = newpath.append
     last = len(segments) - 1
     for pos, seg in enumerate(segments):
-        if seg == '.':
-            seg = ''
-
-        if seg == '':
+        if seg == "" or seg == ".":
+            seg = ""
             allow_empty = (
                 pos == 0
-                or pos == last and newpath and newpath[-1] != ''
-                or pos == last and newpath == ['']
+                or pos == last
+                and newpath
+                and newpath[-1] != ""
+                or pos == last
+                and newpath == [""]
             )
             if not allow_empty:
                 continue
 
-        if seg == '..':
-            if newpath and newpath != ['']:
+        elif seg == "..":
+            if newpath and newpath != [""]:
                 newpath.pop()
             continue
 
-        newpath.append(seg)
+        newpath_append(seg)
 
-    return '/'.join(newpath)
+    return "/".join(newpath)
 
 
-def make_query(data=None, separator=';', charset=None, **kwargs):
+def make_query(data=None, separator=";", charset=None, **kwargs):
     """\
     Return a query string formed from the given dictionary data.
 
@@ -199,19 +209,19 @@ def make_query(data=None, separator=';', charset=None, **kwargs):
     :rtype: str
     """
     if isinstance(data, MultiDict):
-        items = list(data.allitems())
+        item_list = list(data.allitems())
     elif isinstance(data, dict):
-        items = list(data.items())
+        item_list = list(data.items())
     elif data is None:
-        items = []
+        item_list = []
     else:
-        items = list(data)
-    items += list(kwargs.items())
+        item_list = list(data)
+    item_list.extend(kwargs.items())
 
     if charset is None:
         charset = fresco.DEFAULT_CHARSET
 
-    items = _repeat_keys(items)
+    items = _repeat_keys(item_list)
     items = ((k, v) for k, v in items if v is not None)
     return separator.join(_qs_frag(k, v, charset=charset) for k, v in items)
 
@@ -231,13 +241,13 @@ def _qs_frag(key, value, charset=None):
     if charset is None:
         charset = fresco.DEFAULT_CHARSET
 
-    key = ustr(key).encode(charset)
-    value = ustr(value).encode(charset)
+    key = str(key).encode(charset)
+    value = str(value).encode(charset)
 
-    return quote_plus(key) + '=' + quote_plus(value)
+    return quote_plus(key) + "=" + quote_plus(value)
 
 
-def _repeat_keys(iterable):
+def _repeat_keys(iterable: Iterable) -> Iterable:
     """
     Return a list of ``(key, scalar_value)`` tuples given an iterable
     containing ``(key, iterable_or_scalar_value)``.
@@ -255,7 +265,7 @@ def _repeat_keys(iterable):
     """
 
     for key, value in iterable:
-        if isinstance(value, string_types):
+        if isinstance(value, str):
             value = [value]
         else:
             try:
@@ -265,3 +275,51 @@ def _repeat_keys(iterable):
 
         for subvalue in value:
             yield key, subvalue
+
+
+_valid_url_chars = re.compile(
+    r"^(?:[A-Za-z0-9\-._~:/?#\[\]@!$&'\(\)\*\+,;=']+|%[0-9A-Fa-f]{2})*$"
+)
+
+
+def is_safe_url(
+    url,
+    allowed_hosts=frozenset(),
+    allowed_schemes={"http", "https"},
+    valid_url=_valid_url_chars.match,
+    urlparse=urlparse,
+):
+    """
+    Return True if the given URL is for a trusted origin and can safely be
+    redirected to.
+
+    :param allowed_hosts: a list of hosts considered trusted. If not supplied
+                          the current HTTP_HOST and SERVER_NAME are used.
+    :param allowed_schemes: a list of permitted URL schemes. Defaults to
+                            ``{'http', 'https'}``
+    """
+    if not url:
+        return False
+    if url.startswith("///"):
+        return False
+    if not valid_url(url):
+        return False
+
+    if not allowed_hosts:
+        try:
+            from fresco import context
+
+            env = context.request.environ.get
+        except AttributeError:
+            env = {}.get
+        allowed_hosts = {env("HTTP_HOST", env("SERVER_NAME", ""))}
+
+    parsed = urlparse(url)
+    scheme = parsed.scheme
+    netloc = parsed.netloc
+
+    return (
+        (scheme in allowed_schemes and netloc in allowed_hosts)
+        or (scheme == "" and netloc == "")
+        or (scheme == "" and netloc in allowed_hosts)
+    )
