@@ -12,38 +12,97 @@
 #     See the License for the specific language governing permissions and
 #     limitations under the License.
 #
-"""\
+"""
 Utilities for interfacing with WSGI
 """
 
-from __future__ import absolute_import
-
 from functools import partial
 from io import BytesIO
+from urllib.parse import unquote
+from urllib.parse import urlparse
+from typing import List
 import sys
 
-from fresco.compat import unquote, urlparse
-
-__all__ = ['environ_to_unicode',
-           'unicode_to_environ',
-           'environ_to_bytes',
-           'bytes_to_environ',
-           'StartResponseWrapper',
-           'ClosingIterator',
-           'make_environ',
-           ]
+__all__ = [
+    "environ_to_unicode",
+    "unicode_to_environ",
+    "environ_to_str",
+    "str_to_environ",
+    "environ_to_bytes",
+    "bytes_to_environ",
+    "StartResponseWrapper",
+    "ClosingIterator",
+    "make_environ",
+]
 
 
 #: The core keys expected in the WSGI environ dict, as defined by PEP3333
-WSGI_KEYS = set(['REQUEST_METHOD',
-                 'SCRIPT_NAME',
-                 'PATH_INFO',
-                 'QUERY_STRING',
-                 'CONTENT_TYPE',
-                 'CONTENT_LENGTH',
-                 'SERVER_NAME',
-                 'SERVER_PORT',
-                 'SERVER_PROTOCOL'])
+WSGI_KEYS = set(
+    [
+        "REQUEST_METHOD",
+        "SCRIPT_NAME",
+        "PATH_INFO",
+        "QUERY_STRING",
+        "CONTENT_TYPE",
+        "CONTENT_LENGTH",
+        "SERVER_NAME",
+        "SERVER_PORT",
+        "SERVER_PROTOCOL",
+    ]
+)
+
+
+#: List of standard request header field names
+#: Taken from the list available at
+#: http://en.wikipedia.org/wiki/List_of_HTTP_header_fields which combines
+#: RFC specified headers with commonly used non-standard header names
+REQUEST_HEADER_NAMES = {
+    "accept": "Accept",
+    "accept_charset": "Accept-Charset",
+    "accept_encoding": "Accept-Encoding",
+    "accept_language": "Accept-Language",
+    "accept_datetime": "Accept-Datetime",
+    "authorization": "Authorization",
+    "cache_control": "Cache-Control",
+    "connection": "Connection",
+    "cookie": "Cookie",
+    "content_length": "Content-Length",
+    "content_md5": "Content-MD5",
+    "content_type": "Content-Type",
+    "date": "Date",
+    "expect": "Expect",
+    "from": "From",
+    "host": "Host",
+    "if_match": "If-Match",
+    "if_modified_since": "If-Modified-Since",
+    "if_none_match": "If-None-Match",
+    "if_range": "If-Range",
+    "if_unmodified_since": "If-Unmodified-Since",
+    "max_forwards": "Max-Forwards",
+    "origin": "Origin",
+    "pragma": "Pragma",
+    "proxy_authorization": "Proxy-Authorization",
+    "range": "Range",
+    "referer": "Referer",
+    "te": "TE",
+    "user_agent": "User-Agent",
+    "upgrade": "Upgrade",
+    "via": "Via",
+    "warning": "Warning",
+    "common": "Common",
+    "field": "Field",
+    "x_requested_with": "X-Requested-With",
+    "dnt": "DNT",
+    "x_forwarded_for": "X-Forwarded-For",
+    "x_forwarded_host": "X-Forwarded-Host",
+    "x_forwarded_host": "X-Forwarded-Host",
+    "x_forwarded_proto": "X-Forwarded-Proto",
+    "front_end_https": "Front-End-Https",
+    "x_http_method_override": "X-Http-Method-Override",
+    "x_att_deviceid": "X-ATT-DeviceId",
+    "x_wap_profile": "X-Wap-Profile",
+    "proxy_connection": "Proxy-Connection",
+}
 
 
 def with_docstring_from(src):
@@ -53,79 +112,83 @@ def with_docstring_from(src):
         except AttributeError:
             tgt.func_doc = src.func_doc
         return tgt
+
     return with_docstring_from
 
 
-def _environ_to_unicode_py2(s, enc='UTF-8'):
+def environ_to_str(s, enc="UTF-8"):
     """
-    Decode a WSGI environ value to a unicode string
+    Decode a WSGI environ value to a string
     """
-    return s.decode(enc)
+    return s.encode("iso-8859-1").decode(enc)
 
 
-@with_docstring_from(_environ_to_unicode_py2)
-def _environ_to_unicode_py3(s, enc='UTF-8'):
-    return s.encode('iso-8859-1').decode(enc)
-
-
-def _unicode_to_environ_py2(s, enc='UTF-8'):
+def str_to_environ(s, enc="UTF-8"):
     """
-    Return a unicode string encoded for a WSGI environ value
+    Return a string encoded for a WSGI environ value
 
-    For python 2, this simply returns ``s`` encoded using the specified
-    encoding
-
-    For python 3 this returns a 'bytes-as-unicode' string:
+    This returns a 'bytes-as-unicode' string:
 
     - encode ``s`` using the specified encoding (eg utf8)
     - decode the resulting byte string as latin-1
     """
-    return s.encode(enc, 'surrogateescape')
+    return s.encode(enc, "surrogateescape").decode("iso-8859-1")
 
 
-@with_docstring_from(_unicode_to_environ_py2)
-def _unicode_to_environ_py3(s, enc='UTF-8'):
-    return s.encode(enc, 'surrogateescape').decode('iso-8859-1')
+def environ_to_unicode(s, enc="UTF-8"):
+    import warnings
+
+    warnings.warn(
+        "environ_to_unicode is deprecated, use environ_to_str instead",
+        DeprecationWarning,
+    )
+    return environ_to_str(s, enc)
 
 
-def _environ_to_bytes_py2(s):
+def unicode_to_environ(s, enc="UTF-8"):
+    import warnings
+
+    warnings.warn(
+        "unicode_to_environ is deprecated, use environ_to_str instead",
+        DeprecationWarning,
+    )
+    return environ_to_str(s, enc)
+
+
+def environ_to_bytes(s):
     """
     Decode a WSGI environ value to a bytes object
     """
-    return s
+    return s.encode("latin1")
 
 
-@with_docstring_from(_environ_to_bytes_py2)
-def _environ_to_bytes_py3(s):
-    return s.encode('latin1')
-
-
-def _bytes_to_environ_py2(s):
+def bytes_to_environ(s):
     """
     Encode a byte string to a WSGI environ value
 
-    For Python 2, this simply returns ``s``.
-    For Python 3 this returns a 'bytes-as-unicode' string.
+    This returns a 'bytes-as-unicode' string.
     """
-    return s
+    return s.decode("latin1")
 
 
-@with_docstring_from(_environ_to_bytes_py2)
-def _bytes_to_environ_py3(s):
-    return s.decode('latin1')
+def getenv(environ, key, default=None):
+    """
+    Return the named ``key`` from the WSGI environ dict ``environ`` as
+    a byte string.
+    """
+    try:
+        return environ[key].encode("ISO-8859-1")
+    except KeyError:
+        return default
 
 
-if str is bytes:
-    environ_to_unicode = _environ_to_unicode_py2
-    unicode_to_environ = _unicode_to_environ_py2
-    environ_to_bytes = _environ_to_bytes_py2
-    bytes_to_environ = _bytes_to_environ_py2
-
-else:
-    environ_to_unicode = _environ_to_unicode_py3
-    unicode_to_environ = _unicode_to_environ_py3
-    environ_to_bytes = _environ_to_bytes_py3
-    bytes_to_environ = _bytes_to_environ_py3
+def setenv(environ, key, value, enc="UTF-8"):
+    """
+    Set the named ``key`` to ``value`` in the WSGI environ dict ``environ``.
+    Value must be a byte ``str`` and will be encoded according
+    to the rules in PEP-3333.
+    """
+    environ[key] = value.decode("ISO-8859-1")
 
 
 class StartResponseWrapper(object):
@@ -167,7 +230,7 @@ class StartResponseWrapper(object):
     def __init__(self, start_response):
         self.start_response = start_response
         self.status = None
-        self.headers = []
+        self.headers: List[str] = []
         self.called = False
         self.buf = BytesIO()
         self.exc_info = None
@@ -189,9 +252,7 @@ class StartResponseWrapper(object):
         """
         try:
             write = self.start_response(
-                self.status,
-                self.headers,
-                self.exc_info
+                self.status, self.headers, self.exc_info
             )
             write(self.buf.getvalue())
             return write
@@ -228,19 +289,21 @@ class ClosingIterator(object):
 
     """
 
+    __slots__ = ("_iterable", "_next", "_close_funcs", "_closed")
+
     def __init__(self, iterable, *close_funcs):
         """
         Initialize a ``ClosingIterator`` to wrap iterable ``iterable``, and
         call any functions listed in ``*close_funcs`` on the instance's
         ``.close`` method.
         """
-        self.__dict__['_iterable'] = iterable
-        self.__dict__['_next'] = partial(next, iter(self._iterable))
-        self.__dict__['_close_funcs'] = close_funcs
-        iterable_close = getattr(self._iterable, 'close', None)
+        self._iterable = iterable
+        self._next = partial(next, iter(iterable))
+        self._close_funcs = close_funcs
+        iterable_close = getattr(self._iterable, "close", None)
         if iterable_close is not None:
-            self.__dict__['_close_funcs'] = (iterable_close,) + close_funcs
-        self.__dict__['_closed'] = False
+            self._close_funcs = (iterable_close,) + close_funcs
+        self._closed = False
 
     def __iter__(self):
         """
@@ -249,31 +312,19 @@ class ClosingIterator(object):
         return self
 
     def __next__(self):
-        """\
+        """
         Return the next item from the iterator
         """
         return self._next()
 
-    def next(self):
-        """
-        Return the next item from ``iterator``
-        """
-        return self._next()
-
     def close(self):
         """
         Call all close functions listed in ``*close_funcs``.
         """
-        self.__dict__['_closed'] = True
+        self._closed = True
         for func in self._close_funcs:
             func()
 
-    def __getattr__(self, attr):
-        return getattr(self._iterable, attr)
-
-    def __setattr__(self, attr, value):
-        return getattr(self._iterable, attr, value)
-
     def __del__(self):
         """
         Emit a warning if the iterator is deleted without ``close`` having been
@@ -285,43 +336,10 @@ class ClosingIterator(object):
             except ImportError:
                 return
             else:
-                warnings.warn("%r deleted without close being called" %
-                              (self,))
+                warnings.warn("%r deleted without close being called" % (self,))
 
 
-class WSGIStack(object):
-    """
-    A context manager that sets up a WSGI request environ
-    passed through a list of middleware.
-    """
-
-    def __init__(self, environ, middleware_stack):
-        self.environ = environ
-        self.middleware_stack = middleware_stack
-        app = self.fake_app
-        for m in middleware_stack:
-            app = m(app)
-        self.app = app
-        self.content_iterator = self.app(environ, lambda *a, **kw: None)
-
-    def fake_app(self, environ, start_response):
-        start_response('200 OK', [])
-        yield b''
-
-    def __enter__(self):
-        # Invoke the fake WSGI app. This executes any middleware, which may
-        # have side effects required downstream (eg setting environ keys)
-        next(self.content_iterator, None)
-        return self.environ
-
-    def __exit__(self, type, value, traceback):
-        # Exhaust the iterator and close up
-        list(self.content_iterator)
-        getattr(self.content_iterator, 'close', lambda: None)()
-
-
-def make_environ(url='/', environ=None,
-                    wsgi_input=b'', middleware=True, **kwargs):
+def make_environ(url="/", environ=None, wsgi_input=b"", **kwargs):
     """
     Return a WSGI environ dict populated with values modelling the specified
     request url and data.
@@ -332,67 +350,62 @@ def make_environ(url='/', environ=None,
     :param environ: values to pass into the WSGI environ dict
     :param wsgi_input: The input stream to use in the ``wsgi.input``
                         key of the environ dict
-    :param middleware: If ``False`` the middleware stack will not be
-                        invoked. Disabling the middleware can speed up
-                        the execution considerably, but it will no longer
-                        give an accurate simulation of a real HTTP request.
     :param kwargs: additional keyword arguments will be passed into the
                     WSGI request environment
     """
-    from fresco.request import HEADER_NAMES as REQUEST_HEADER_NAMES
     url = urlparse(url)
     netloc = url.netloc
-    user = ''
-    if '@' in netloc:
-        user, netloc = netloc.split('@', 1)
+    user = ""
+    if "@" in netloc:
+        user, netloc = netloc.split("@", 1)
 
-    if ':' in user:
-        user, _ = user.split(':')[0]
+    if ":" in user:
+        user = user.split(":")[0]
 
     if isinstance(wsgi_input, bytes):
         wsgi_input = BytesIO(wsgi_input)
 
     env_overrides = environ or {}
     for key, value in kwargs.items():
-        key = key.replace('wsgi_', 'wsgi.')
-        if '.' not in key:
+        key = key.replace("wsgi_", "wsgi.")
+        if "." not in key:
             # Convert core WSGI keys to upper case
             if key.upper() in WSGI_KEYS:
                 key = key.upper()
             # Convert header names to form HTTP_USER_AGENT
             elif key.lower() in REQUEST_HEADER_NAMES:
-                key = 'HTTP_' + key.upper()
+                key = "HTTP_" + key.upper()
             # value must be a python native str, whatever that means
             if not isinstance(value, str):
-                value = unicode_to_environ(value)
+                value = str_to_environ(value)
         env_overrides[key] = value
 
     environ = {
-        'REQUEST_METHOD': 'GET',
-        'SERVER_NAME': 'localhost',
-        'SERVER_PORT': '80',
-        'SERVER_PROTOCOL': 'HTTP/1.0',
-        'HTTP_HOST': netloc or 'localhost',
-        'SCRIPT_NAME': '',
-        'PATH_INFO': unicode_to_environ(unquote(url.path)),
-        'REMOTE_ADDR': '127.0.0.1',
-        'wsgi.version': (1, 0),
-        'wsgi.url_scheme': url.scheme or 'http',
-        'wsgi.input': wsgi_input,
-        'wsgi.errors': sys.stderr,
-        'wsgi.multithread': True,
-        'wsgi.multiprocess': True,
-        'wsgi.run_once': False,
+        "REQUEST_METHOD": "GET",
+        "SERVER_NAME": "localhost",
+        "SERVER_PORT": "80",
+        "SERVER_PROTOCOL": "HTTP/1.0",
+        "HTTP_HOST": netloc or "localhost",
+        "SCRIPT_NAME": "",
+        "PATH_INFO": str_to_environ(unquote(url.path)),
+        "REMOTE_ADDR": "127.0.0.1",
+        "wsgi.version": (1, 0),
+        "wsgi.url_scheme": url.scheme or "http",
+        "wsgi.input": wsgi_input,
+        "wsgi.errors": sys.stderr,
+        "wsgi.multithread": True,
+        "wsgi.multiprocess": True,
+        "wsgi.run_once": False,
     }
-    if url.scheme == 'https':
-        environ['HTTPS'] = 'on'
-        environ['SERVER_PORT'] = '443'
+    if url.scheme == "https":
+        environ["HTTPS"] = "on"
+        environ["SERVER_PORT"] = "443"
 
     if user:
-        environ['REMOTE_USER'] = user
+        environ["REMOTE_USER"] = user
 
     if url.query:
-        environ['QUERY_STRING'] = url.query
+        environ["QUERY_STRING"] = url.query
 
     environ.update(env_overrides)
     return environ
